/**
 * Appcelerator Proprietary and Confidential
 */
 
#import "TiCompiledMethod.h"

static NSMutableArray *callstack = nil;
static NSRecursiveLock *stacklock = nil;


@implementation TiCompiledMethod


-(id)initWithDelegate:(KrollMethod*)method args:(id)args_ code:(NSString*)code_ source:(NSString*)source_ line:(NSNumber*)line_ column:(NSNumber*)column_
{
    if ((self = [super initWithTarget:[method target] selector:[method selector] argcount:[method argcount] type:[method type] name:[method name] context:[method context]]))
    {
        delegate = [method retain];
        args = [args_ retain];
        source = [source_ copy];
        code = [code_ copy];
        line = [line_ intValue];
        column = [column_ intValue];
    }
    return self;
}

-(void)dealloc
{
    RELEASE_TO_NIL(source);
    RELEASE_TO_NIL(code);
    RELEASE_TO_NIL(delegate);
    RELEASE_TO_NIL(args);
    [super dealloc];
}

-(NSString*)callstack
{
    return [NSString stringWithFormat:@"%@ (%@:%d)", code, source, line];
}

+(NSString*)stack
{
    NSMutableString *trace = [NSMutableString string];
    
    [stacklock lock];
   
    for (TiCompiledMethod *m in callstack)
    {
        [trace appendFormat:@"     at %@\n", [m callstack]];
    }
    
    [stacklock unlock];
    
    return trace;
}

-(id)call:(NSArray*)args_
{
    if (stacklock==nil)
    {
        stacklock = [[NSRecursiveLock alloc] init];
        callstack = [[NSMutableArray alloc] init];
    }
    
    // we'll need to lock to use stack traces.  maybe this could be configurable or only in a certain build mode
    
    [stacklock lock];
    [callstack addObject:self];
    [stacklock unlock];

    if (args!=nil)
    {
        args_ = args;
    }
    
    id result = [super call:args_];

    [stacklock lock];
    [callstack removeObject:self];
    [stacklock unlock];
    
    return result;
}

/*
 without locking:
 
 [INFO] Ti.API.log = 218 ms
 console.log = 172 ms
 compiled = 113 ms
 
 with locking:
 
 Ti.API.log = 220 ms
 console.log = 180 ms
 compiled = 130 ms
*/

@end
