/**
 * Appcelerator Proprietary and Confidential
 */
 
#import "TiCompilerModule.h"
#import "TiBase.h"
#import "TiHost.h"
#import "TiUtils.h"
#import "KrollBridge.h"
#import "KrollCallback.h"
#import "KrollObject.h"
#import "KrollContext.h"
#import "KrollMethod.h"
#import "TiCompiledMethod.h"

@implementation TiCompilerModule

#pragma mark Internal

// this is generated for your module, please do not change it
-(id)moduleGUID
{
	return @"04a86c45-7131-4b98-9fa1-9de441396c8b";
}

// this is generated for your module, please do not change it
-(NSString*)moduleId
{
	return @"ti.compiler";
}

#pragma crash delegate

- (void)handleUncaughtException:(NSException *)exception withStackTrace:(NSArray *)stackTrace
{
	// get the app stack trace
    NSString *stacktrace = [TiCompiledMethod stack];
    NSString *message = [NSString stringWithFormat:@"%@\n%@", [exception description], stacktrace];
    NSLog(@"[ERROR] %@",message);

    TiThreadPerformOnMainThread(^{
    
        UIAlertView *alert = [[UIAlertView alloc] initWithTitle:@"Unhandled Exception"
                                                        message:message
                                                     delegate:nil
                                            cancelButtonTitle:@"OK"
                                            otherButtonTitles:nil];
        [alert show];
    },YES);
    
}

- (void)handleScriptError:(TiScriptError *)error
{
	// get the app stack trace
    NSString *stacktrace = [TiCompiledMethod stack];
    NSString *message = [NSString stringWithFormat:@"%@\n%@", [error message], stacktrace];
    NSLog(@"[ERROR] %@",message);
    
    //TODO: gather env details & send analytics
    //TODO: quit app

    TiThreadPerformOnMainThread(^{
    
        UIAlertView *alert = [[UIAlertView alloc] initWithTitle:@"Script Error"
                                                          message:message
                                                         delegate:nil
                                                cancelButtonTitle:@"OK"
                                                otherButtonTitles:nil];
        [alert show];
    },YES);
}


#pragma Implementation

-(void)crash:(id)args
{
    //SCRIPT ERROR: [[NSArray array] objectAtIndex:0];
    abort();
}

-(void)compile:(id)_args
{
    
    //TODO: do we need to remember proxy stuff -- ask Blain
    

    //setup our exception handler
    //TODO: we should potentially re-delegate to existing delegate
    [TiExceptionHandler defaultExceptionHandler].delegate = self;
    
    
    //setup our optimized functions
    TiHost *_host = [self _host];
    id<TiEvaluator> ctx = [self pageContext];
    KrollContext *kctx = [ctx krollContext];
	TiGlobalContextRef jsContext = [kctx context]; 
	TiObjectRef globalRef = TiContextGetGlobalObject(jsContext);
    
    NSArray *symbolTable = nil;
     
    for (NSArray *arg in _args)
    {
    	// the first argument is the symbol table (which is an array of symbol entries that will be later referenced by their index)
        if (symbolTable==nil)
        {
            symbolTable = arg;
            continue;
        }
        
        NSString *moduleName = [symbolTable objectAtIndex:[[arg objectAtIndex:0] intValue]];
        NSString *methodName = [symbolTable objectAtIndex:[[arg objectAtIndex:1] intValue]];
        NSString *source = [symbolTable objectAtIndex:[[arg objectAtIndex:2] intValue]];
        NSNumber *line = [arg objectAtIndex:3];
        NSNumber *column = [arg objectAtIndex:4];
        NSString *symbol = [arg objectAtIndex:5];
        id args = [arg objectAtIndex:6];
        
        if ([args isEqual:[NSNull null]] || [args count] == 0)
        {
            args = nil;
        }
        else
        {
            NSMutableArray *a = [NSMutableArray arrayWithCapacity:[args count]];
            for (NSNumber *idx in args)
            {
                id value = [symbolTable objectAtIndex:[idx intValue]];
                [a addObject:value];
            }
            args = a;
        }
        
        TiModule *mod = [_host moduleNamed:moduleName context:ctx];
        KrollObject *ko = [[KrollObject alloc] initWithTarget:mod context:kctx];
        KrollMethod *m = (KrollMethod*)[ko valueForKey:methodName];
       
        NSString *code = [NSString stringWithFormat:@"Ti.%@.%@",moduleName,methodName];
        
        // create the compiled method delegate
        TiCompiledMethod *cm = [[TiCompiledMethod alloc] initWithDelegate:m args:args code:code source:source line:line column:column];

        
        TiStringRef prop = TiStringCreateWithCFString((CFStringRef)symbol);
        
        // bind a new symbol in the compiler with a pointer to the bound compiled method
        // right now, will still have to create a TiCompiledMethod for each symbol we've compiled (which we will share the same function/symbol)
        // however, it would be nice in the future, if we could actually use a real C function pointer so that we can eliminate the extra object
        // overhead (memory, processing, garbage) as well as gain a real jump table like functionality. for now, this is still faster than the 
        // current alternative.  also, as an interimediate step, we could use a compiler plug in to generate the C function routine pointer and 
        // dynamically generate that function for each symbol and avoid the TiCompiledMethod object -- however, that wouldn't be as easy and quick
        // as a module.
        TiObjectSetProperty(jsContext, globalRef, prop, [KrollObject toValue:kctx value:cm], kTiPropertyAttributeReadOnly|kTiPropertyAttributeDontDelete, NULL);
        TiStringRelease(prop);

        [cm release];
        [ko release];
    }
}



@end
