/**
 * Titanium Optimizing Compiler Prototype
 * Copyright (c) 2012 by Appcelerator, Inc.
 * Proprietary & Confidential - Do not redistribute
 */

const SYMBOL_PREFIX = 'T$';

var path = require('path'),
	fs = require('fs'),
	util = require('util'),
	jsp = require("uglify-js").parser,
	pro = require("uglify-js").uglify,
	isTiMethod = /^Ti(tanium)?\./,
	symbolCount = 1,
	symbolTable = [],
	apis = [],
	isJSFile = /.js$/,
	afs,
	i18n,
	__,
	__n,
	logger;
	

//
// TODO: optimizations
//
// rollout subsequent properties on createWindow (etc).
//
// var w = Ti.UI.createWindow();
// w.width = Ti.UI.FILL;
// w.height = Ti.UI.FILL;
//
// would compile into:
//
// var w = Ti.UI.createWindow({width:Ti.UI.FILL,height:Ti.UI.FILL});
//
// we could also consider:
//
// w.add(view1);
// w.add(view2);
// w.add(view3);
//
// would compile into:
//
// w.add(view1,view2,view3);
//
//

function transformConstants(code)
{
	//TODO: this is just a few of them, we should figure out how to do them all (from jsca)
	return code.replace(/Titanium\./g,'Ti.')
			   .replace(/Ti\.UI\.FILL/g,'"FILL"')
			   .replace(/Ti\.UI\.SIZE/g,'"SIZE"')
			   .replace(/Ti\.UI\.AUTO/g,'"auto"');
}

function addArgument(arg)
{
	for (var c=0;c<symbolTable.length;c++)
	{
		if (arg===symbolTable[c])
		{
			return c;
		}
		if (typeof(arg)=='object' && JSON.stringify(arg)==JSON.stringify(symbolTable[c]))
		{
			return c;
		}
	}
	symbolTable.push(arg);
	return symbolTable.length-1;
}

function stringifyObj(arg)
{
	if (arg[0]=='object')
	{
		var o = {};
		for (var x=0;x<arg[1].length;x++)
		{
			var xo = arg[1][x];
			o[xo[0]] = stringifyObj(xo[1]);
		}
		return o;
	}
	else if (arg[0]=='name')
	{
		if (arg[1]=='true')
			return true;
		if (arg[1]=='false')
			return false;
	}
	else if (arg[0]=='function')
	{
		return pro.gen_code(arg);
	}
	return arg[1];
}

function addAPI(module, name, source, line, column, symbol, args)
{
	var a = args ? [] : null;
	if (args)
	{
		for (var c=0;c<args.length;c++)
		{
			a.push(addArgument(stringifyObj(args[c])))
		}
	}
	apis.push([
		addArgument(module),
		addArgument(name),
		addArgument(source),
		line,
		column,
		symbol,
		a
	]);
}

function getAPIName(segment) 
{
	if (segment)
	{
		switch(segment[0])
		{
			case 'dot':
				return getAPIName(segment[1]).concat(segment[2]);
		    case 'name':
		    	return [segment[1]];
		    default:
				return [''];
		}
	}
}

function isTiModule(name)
{
	//TODO: for now, this is hardcoded here until we move this into Bryan's new code processor which reads the JSCA
	return /^Ti(tanium)?\.(API|Accelerometer|API|App|Analytics|Contacts|Codec|Database|Facebook|Filesystem|Geolocation|Gesture|Locale|Map|Media|Network|Platform|Stream|Utils|UI|XML|Yahoo|)\./.test(name);
}

function generateAPISymbol()
{
	return SYMBOL_PREFIX + symbolCount++;
}

function isStaticHash(obj)
{
	for (var c=0;c<obj.length;c++)
	{
		if (!isTypeStatic(obj[c][1]))
		{
			return false;
		}
	}
	return true;
}

function isTypeStatic(obj)
{
	var name = obj[0].name;
	if (!/(string|num)/.test(name))
	{
		if (name == 'name' && (obj[1]==='true' || obj[1]==='false'))
		{
			// this is a boolean value, it's OK
			return true;
		}
		if (name === 'object' && isStaticHash(obj[1]))
		{
			// this is a static hash
			return true;
		}
		if (name === 'function')
		{
			return true;
		}
		return false;
	}
	return true;
}

function isArgsStatic(args)
{
	// check to see if all the arguments are static (can be safely passed in compiled method)
	if (args.length > 0)
	{
		for (var c=0;c<args.length;c++)
		{
			if (!isTypeStatic(args[c]))
			{
				return false;
			}
		}
		return true;
	}
	return false;
}

function instrument(code,filename) 
{
	code = transformConstants(code);


	function transformAPI(name,args,body) 
	{
		var ret;
		if (this[0].start) 
		{
			var apiName = getAPIName(this[1]).join('.');

			if (isTiMethod.test(apiName) && isTiModule(apiName))
			{
				var i = apiName.lastIndexOf('.'),
					cn = apiName.substring(0,i).replace('Titanium','Ti').replace('Ti.','').replace('.',''),
					m = apiName.substring(i+1),
					staticArgs = isArgsStatic(args),
					symbol = generateAPISymbol();
					
					
				addAPI(cn, m, filename, this[0].start.line+1, this[0].start.col+1, symbol, staticArgs ? args : null);

				return [ "call", [ "name", symbol ], staticArgs ? null : args ];
			}
		}
		return ret;
	}

	// for now, this only optimizes function calls

	var ast = jsp.parse(code, false, true); 
	var w = pro.ast_walker();
	var new_ast = w.with_walkers({
		"call" : transformAPI,
	}, function(){
		return w.walk(ast);
	});

	return pro.gen_code(new_ast, { beautify: true });
}

function stringifyApis(obj)
{
	var str = '[';
	for (var c=0;c<obj.length;c++)
	{
		var v = obj[c];
		switch(typeof(v))
		{
			case 'string':
				if (/^\(function\(/.test(v))
				{
					str+=v;
				}
				else
				{
					str+='"' + v + '"';
				}
				break;
			case 'number':
			case 'boolean':
				str+=v;
				break;
			case 'object':
			default:
				str+=JSON.stringify(v);
				break;
		}
		if (c + 1 < obj.length)
		{
			str+=',';
		}
	}
	str+=']';
	return str;
}

function compileFile(filename, fullpath)
{
	// only need to compile JS generated files
	if (isJSFile.test(filename))
	{
		logger.debug('[optimizer]'.yellow +' '+__('compiling')+' '+fullpath.cyan);
		
		var contents = fs.readFileSync(fullpath,"utf-8");
    	var code = instrument(contents, filename);

	    if (apis.length>0)
	    {
			var js = ('require("ti.compiler").compile(\n');
			// first argument is always the symbol table
			js+="   "+stringifyApis(symbolTable)+',\n';
			// then subsequent arguments are the apis
			for (var c=0;c<apis.length;c++)
			{
				js+="   "+(JSON.stringify(apis[c]) + (c +1 < apis.length ? "," : "")) + "\n";
			}
			js+=(');\n');
			js+=code;
			fs.writeFileSync(fullpath, js, 'utf-8');
			logger.info('[optimizer]'.yellow + ' '+__('compiled')+' '+fullpath.green);
	    }
	    else
	    {
			logger.debug('[optimizer]'.yellow +' '+__('skipped')+' '+fullpath.grey);
	    }

	    // reset the globals
		symbolCount = 1;
		symbolTable = [];
		apis = [];
	}
}

//--------------------------------------------------------------------------------//
// CLI
//--------------------------------------------------------------------------------//
exports.cliVersion = '>=3.X';

exports.init = function (log, config, cli, appc) 
{
	// the ti.compiler hook must come after the ti.alloy hook

	cli.addHook('build.pre.compile', function (build, finished) 
	{
		afs = appc.fs;
		i18n = appc.i18n(__dirname);
		__ = i18n.__;
		__n = i18n.__n;
		logger = log;


		/**  TODO: right now, we're doing it all the time but you'd normally want to do it only on non dev
		if (build.deployType === 'development')
		{
			logger.info(__('Skipped optimizer in non-production mode'));
			return finished();
		}
		**/

		var alloyDir = path.join(cli.argv['project-dir'], 'app');

		if (!afs.exists(alloyDir))
		{
			logger.info(__('Skipping optimizing compiler, not an Alloy project'));
			return finished();
		}

		var resourcesDir = path.join(cli.argv['project-dir'], 'Resources');

		if (!afs.exists(resourcesDir)) {
			logger.error(__('No Resources directory found!'));
			finished();
			return;
		}

		afs.visitFiles(resourcesDir,compileFile,finished);
	});
};

